////////////////////////////////////////////////////////
//
// GEM - Graphics Environment for Multimedia
//
// zmoelnig@iem.kug.ac.at
//
// Implementation file
//
//    Copyright (c) 1997-1999 Mark Danks.
//    Copyright (c) Gnther Geiger.
//    Copyright (c) 2001-2002 IOhannes m zmoelnig. forum::fr::umlute. IEM
//    Copyright (c) 2005 Pierre-Olivier Charlebois McGill Electrical Engineering
//    
//	  For information on usage and redistribution, and for a DISCLAIMER OF ALL
//    WARRANTIES, see the file, "GEM.LICENSE.TERMS" in this distribution.
//
/////////////////////////////////////////////////////////

#include "spot_light.h"

#include "Base/GemMan.h"

#ifdef __APPLE__
extern bool HaveValidContext (void);
#endif

CPPEXTERN_NEW_WITH_ONE_ARG(spot_light, t_floatarg, A_DEFFLOAT)

/////////////////////////////////////////////////////////
//
// spot_light
//
/////////////////////////////////////////////////////////
// Constructor
//
/////////////////////////////////////////////////////////

spot_light :: spot_light(t_floatarg lightNum)
			 : m_debug(0), m_thing(NULL)
{
    // initialize to white
	m_color[0] = m_color[1] = m_color[2] = m_color[3] = 1.0;

    // position
	m_position[0] = 0.0;
    m_position[1] = 0.0;
    m_position[2] = 0.0;
    m_position[3] = 1.0;

	constantAttenuation = 0.0;
	linearAttenuation = 0.1;
	quadraticAttenuation = 0.0;
	spotCutoff = 30.0;
	spotExponent = 2.5;

	int num;
	if (lightNum < 1.f)
		num = 0;
	else
		num = (int)lightNum;
	m_light = GemMan::requestLight(num);
	m_on = 1;
	m_change = 1;
	    
	// create the parameters inlet
    inlet_new(this->x_obj, &this->x_obj->ob_pd, gensym("list"), gensym("paramlist"));

    // create the color inlet
    inlet_new(this->x_obj, &this->x_obj->ob_pd, gensym("list"), gensym("clrlist"));

    
}

/////////////////////////////////////////////////////////
// Destructor
//
/////////////////////////////////////////////////////////
spot_light :: ~spot_light()
{
#ifdef __APPLE__
    if ( !HaveValidContext () ){
	if (m_light)
	    GemMan::freeLight(m_light);
	return;
    }
#endif
    stopRendering();
    if (m_light)
	GemMan::freeLight(m_light);
}

/////////////////////////////////////////////////////////
// lightOnOffMess
//
/////////////////////////////////////////////////////////
void spot_light :: lightOnOffMess(int state)
{
    m_on = state;
	m_change = 1;
    setModified();
}

/////////////////////////////////////////////////////////
// debugMess
//
/////////////////////////////////////////////////////////
void spot_light :: debugMess(int state)
{
    m_debug = state;
	m_change = 1;
    setModified();
}

/////////////////////////////////////////////////////////
// lightColorMess
//
/////////////////////////////////////////////////////////
void spot_light :: lightColorMess(float red, float green, float blue)
{
    m_color[0] = red;
    m_color[1] = green;
    m_color[2] = blue;
	m_change = 1;
    setModified();
}

/////////////////////////////////////////////////////////
// lightParamMess
//
/////////////////////////////////////////////////////////

void spot_light :: lightParamMess(float linAtt, float cutoff, float exponent)
{

    // convert from spherical coordinates

    linearAttenuation = linAtt;
    spotCutoff = cutoff;
    spotExponent = exponent;
    	m_change = 1;
    setModified();
}

/////////////////////////////////////////////////////////
// startRendering
//
/////////////////////////////////////////////////////////
void spot_light :: startRendering()
{
    if (m_thing)
		stopRendering();
    m_thing = gluNewQuadric();
	gluQuadricTexture(m_thing, GL_FALSE);
	gluQuadricDrawStyle(m_thing, (GLenum) GLU_FILL);
	m_change = 1;
}

/////////////////////////////////////////////////////////
// stopRendering
//
/////////////////////////////////////////////////////////
void spot_light :: stopRendering()
{
#ifdef __APPLE__
  if (!HaveValidContext ()) {
    post("GEM: spot_light: need context to call glDisable");
    return;
  }
#endif
    if (m_thing)
		gluDeleteQuadric(m_thing);
    m_thing = NULL;
	if (m_light)
		glDisable(m_light);
	m_change = 1;
}

/////////////////////////////////////////////////////////
// render
//
/////////////////////////////////////////////////////////
void spot_light :: render(GemState *state)
{
    if (!m_light)
		return;

	if (m_change)
	{
		m_change = 0;
		if ( !m_on )
		{
    		glDisable(m_light);
    		return;
		}

		
		glEnable(m_light);		

		// setGlobal parameter
		glLightfv(m_light, GL_DIFFUSE,  m_color);
		glLightfv(m_light, GL_SPECULAR, m_color);
		glLightf(m_light, GL_LINEAR_ATTENUATION, linearAttenuation);
		
	}

	if (m_on)
	{
		

		// Set spot parameters
		glLightfv(m_light, GL_POSITION, spotPosition);
		glLightf(m_light, GL_SPOT_EXPONENT, spotExponent);
		glLightfv(m_light, GL_SPOT_DIRECTION, spotDirection);
		glLightf(m_light, GL_SPOT_CUTOFF, spotCutoff);
		
		if (m_debug) // display source
		{
			
			glDisable(GL_LIGHTING);
				
			glColor3fv(m_color);
			gluCylinder(m_thing, 0.2f, 0, 0.4f, 10, 10);
				
			glEnable(GL_LIGHTING);

		}
	}
}

/////////////////////////////////////////////////////////
// static member functions
//
/////////////////////////////////////////////////////////
void spot_light :: obj_setupCallback(t_class *classPtr)
{
	class_addmethod(classPtr, (t_method)&spot_light::lightParamMessCallback,
    	    gensym("paramlist"), A_FLOAT, A_FLOAT, A_FLOAT, A_NULL);
    class_addmethod(classPtr, (t_method)&spot_light::lightColorMessCallback,
    	    gensym("clrlist"), A_FLOAT, A_FLOAT, A_FLOAT, A_NULL);
    
    	class_addmethod(classPtr, (t_method)&spot_light::debugMessCallback,
    	    gensym("debug"), A_FLOAT, A_NULL);
    	class_addfloat(classPtr, (t_method)spot_light::lightOnOffMessCallback);
}
void spot_light :: lightColorMessCallback(void *data, t_floatarg red, t_floatarg green, t_floatarg blue)
{
    GetMyClass(data)->lightColorMess((float)red, (float)green, (float)blue);
}

void spot_light :: lightParamMessCallback(void *data, t_floatarg linAtt, t_floatarg cutoff, t_floatarg exponent)
{
    GetMyClass(data)->lightParamMess((float)linAtt, (float)cutoff, (float)exponent);
}
void spot_light :: lightOnOffMessCallback(void *data, t_floatarg n)
{
    GetMyClass(data)->lightOnOffMess((int)n);
}
void spot_light :: debugMessCallback(void *data, t_floatarg n)
{
    GetMyClass(data)->debugMess((int)n);
}
